#include <Arduino.h>

#include "wifi_client.h"
#include "common.h"

#include <WiFi.h>
#include <esp_wifi.h>

int WiFiScanGetStrongestBSSID(bool do_scan);
void WifiStaConnect(int mode);
void WiFiStaSetConfig(int mode);

#define CHECKBYTE 0b10101010
static RTC_DATA_ATTR struct {
  byte mac[6];
  byte mode;
  int32_t channel;
  IPAddress ip;
  IPAddress gw;
  IPAddress mask;
  IPAddress dns;
  byte chk;
} cfgbuf;

bool CheckCfg() {
  if (cfgbuf.chk == CHECKBYTE) return true;

  // bad checkbyte, init data
  for (uint32_t i = 0; i < 6; i++) cfgbuf.mac[i] = 0xff;
  cfgbuf.mode = 0;
  cfgbuf.channel = 0;
  cfgbuf.ip = IPAddress(0, 0, 0, 0);
  cfgbuf.gw = IPAddress(0, 0, 0, 0);
  cfgbuf.mask = IPAddress(255, 255, 255, 0);
  cfgbuf.dns = IPAddress(0, 0, 0, 0);
  cfgbuf.chk = CHECKBYTE;
  return false;
}

void WriteCfg(void) {
  // save new info
  uint8_t *bssid = WiFi.BSSID();
  for (int i=0; i<sizeof(cfgbuf.mac); i++) cfgbuf.mac[i] = bssid[i];
  cfgbuf.channel = WiFi.channel();
  cfgbuf.ip = WiFi.localIP();
  cfgbuf.gw = WiFi.gatewayIP();
  cfgbuf.mask = WiFi.subnetMask();
  cfgbuf.dns = WiFi.dnsIP();
  cfgbuf.chk = CHECKBYTE;
}


void WiFiInitSta(void){ // returns operating mode
  CheckCfg();

  if((wakeups % 48) == 1){ // every 48 wakeups (4 hours), do a full wifi scan
    cfgbuf.mode = 0;
  }
  
  switch (cfgbuf.mode) {
    default:
    case 0:
      WifiStaConnect(0); // do full scan & connect
      break;
    case 1:
      WifiStaConnect(1); // do quick connect
      break;
  }

  wifi_mode = cfgbuf.mode;

  // wait for connection
  unsigned long start=millis();
  while (!WiFi.isConnected() && (millis() - start < WIFI_CONNECT_TIMEOUT*1e3)){
    delay(10);
  }

  delay(100); // wait to ensure wifi connection is stable

  if(cfgbuf.mode == 0 && WiFi.isConnected()){
    DebugPrintln("WiFi: connected, saving config");
    cfgbuf.mode = 1; // store settings for next deep-sleep boot
    WriteCfg();

  }else if(cfgbuf.mode == 1 && !WiFi.isConnected()){
    DebugPrintln("WiFi: failed to do quickconnect, doing full scan");
    cfgbuf.mode = 0;
    WiFi.disconnect();
    WiFi.mode(WIFI_OFF);
    WiFiInitSta();

  }else if(!WiFi.isConnected()){
    DebugPrintln("WiFi: failed to connect");
  }
}

void WifiStaConnect(int mode){
  // attempt reconnection
  // mode=0 normal init, scan and connect
  // mode=1 quick init and connect using stored cfgbuf values

  DebugPrintln("WiFi: attempting to connect...");

  switch(mode){
    case 0:
    default:
      {
        WiFiStaSetConfig(0);
        
        int i_strongest = WiFiScanGetStrongestBSSID(true);
        if(i_strongest < 0){
          DebugPrintln("WiFi: scan error, access point not found");
          return;
        }
        WiFi.begin(ModuleConfig.wifi_ssid.c_str(), ModuleConfig.wifi_password.c_str(), WiFi.channel(i_strongest), WiFi.BSSID(i_strongest));
      }
      break;

    case 1:
      WiFiStaSetConfig(1);
      //WiFi.config(cfgbuf.ip, cfgbuf.gw, cfgbuf.mask, cfgbuf.dns); // use previous IP
      WiFi.begin(ModuleConfig.wifi_ssid, ModuleConfig.wifi_password, cfgbuf.channel, cfgbuf.mac); // connect to stored config
      break;
  }
}


void WiFiStaSetConfig(int mode){
  // mode 0 is normal init, mode 1 is quick init
  WiFi.mode(WIFI_STA);

  if(mode == 0){
    WiFi.persistent(false);
    if(wakeups==1) WiFi.disconnect(false,true); // remove any stored wifi config from NVS, and disconnect in case there was a wifi config in NVS
    WiFi.setScanMethod(WIFI_ALL_CHANNEL_SCAN);
    WiFi.setSortMethod(WIFI_CONNECT_AP_BY_SIGNAL);
  }
  
  WiFi.setHostname(ModuleConfig.node_name.c_str());
  tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_STA, ModuleConfig.node_name.c_str());
}

int WiFiScanGetStrongestBSSID(bool do_scan){
  int n = 0;
  int i_strongest = -1;
  int32_t rssi_strongest = -120;

  if(do_scan){
    n = WiFi.scanNetworks();
  }else{
    n = WiFi.scanComplete();
  }

  if(n >= 0){
    for (int i = 0; i < n; ++i) {
      // find strongest AP with correct SSID
      if ((String(ModuleConfig.wifi_ssid.c_str()) == String(WiFi.SSID(i)) && (WiFi.RSSI(i)) > rssi_strongest)) {
        rssi_strongest = WiFi.RSSI(i);
        i_strongest = i;
      }
    }
  }else{
    i_strongest=-1;
  }

  return(i_strongest);
}
